# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

if sys.platform == "linux":
    import fcntl

    def test_unpatched_attributes_are_forwarded_to_real_fs(fs):
        # regression test for #1074
        with open("lock_file", "a+") as lock_file:
            fcntl.flock(lock_file, fcntl.LOCK_SH)
            fcntl.flock(lock_file, fcntl.LOCK_UN)
