# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path

import pytest


# Used for testing paused patching during reporting.


@pytest.hookimpl
def pytest_runtest_logreport(report):
    if report.when == "call":
        report_path = Path(__file__).parent / "report.txt"
        with open(report_path, "w") as f:
            f.write("Test")
